import os 
import sys
import uuid
import tqdm
import json
import traceback
from pathlib import Path
from typing import Callable

from openfactcheck.lib.logger import logger
from openfactcheck.lib.config import OpenFactCheckConfig
from openfactcheck.core.solver import SOLVER_REGISTRY, Solver
from openfactcheck.core.state import FactCheckerState

class OpenFactCheck:
    """
    OpenFactCheck class to evaluate the factuality of a response using a pipeline of solvers.

    Parameters
    ----------
    config : OpenFactCheckConfig
        An instance of OpenFactCheckConfig containing the configuration
        settings for OpenFactCheck.

    Attributes
    ----------
    logger : Logger
        An instance of the logger to log messages.
    config : OpenFactCheckConfig
        An instance of OpenFactCheckConfig containing the configuration
        settings for OpenFactCheck.
    solver_configs : dict
        A dictionary containing the configuration settings for the solvers.
    pipeline : list
        A list of solvers to be included in the pipeline.
    output_path : str
        The path to the output directory where the results will be stored.

    Methods
    -------
    load_solvers(solver_paths)
        Load solvers from the given paths.
    list_solvers()
        List all registered solvers.
    list_claimprocessors()
        List all registered claim processors.
    list_retrievers()
        List all registered retrievers.
    list_verifiers()
        List all registered verifiers.
    init_solver(solver_name, args)
        Initialize a solver with the given configuration.
    init_solvers()
        Initialize all registered solvers.
    init_pipeline()
        Initialize the pipeline with the given configuration.
    init_pipeline_manually(pipeline)
        Initialize the pipeline with the given configuration.
    persist_output(state, idx, solver_name, cont, sample_name=0)
        Persist the output of the solver.
    read_output(sample_name)
        Read the output file for the given sample.
    remove_output(sample_name)
        Remove the output file for the given sample.
    __call__(response, question, callback_fun, **kwargs)
        Evaluate the response using the pipeline.

    Examples
    --------
    >>> config = OpenFactCheckConfig("config.json")
    >>> ofc = OpenFactCheck(config)
    >>> response, sample_name = ofc("This is a sample response.")
    >>> output = ofc.read_output(sample_name)
    >>> ofc.remove_output(sample_name)
    """
    def __init__(self, config: OpenFactCheckConfig):
        """
        Initialize OpenFactCheck with the given configuration.

        Parameters
        ----------
        config : OpenFactCheckConfig
            An instance of OpenFactCheckConfig containing the configuration
            settings for OpenFactCheck.
        """
        self.logger = logger
        self.config = config    

        # Initialize attributes
        self.solver_configs = self.config.solver_configs
        self.pipeline = self.config.pipeline
        self.output_path = os.path.abspath(self.config.output_path)

        # Load and register solvers
        self.load_solvers(self.config.solver_paths)
        self.logger.info(f"Loaded solvers: {list(self.list_solvers().keys())}")

        # Initialize the pipeline
        self.pipeline = self.init_pipeline()

        self.logger.info("-------------- OpenFactCheck Initialized ----------------")
        self.logger.info("Pipeline:")
        for idx, (name, (solver, iname, oname)) in enumerate(self.pipeline.items()):
            self.logger.info(f"{idx}-{name} ({iname} -> {oname})")
        self.logger.info("---------------------------------------------------------")

    @staticmethod
    def load_solvers(solver_paths):
        """
        Load solvers from the given paths
        """
        for solver_path in solver_paths:
            abs_path = Path(solver_path).resolve()
            if abs_path.is_dir():
                sys.path.append(str(abs_path.parent))
                Solver.load(str(abs_path), abs_path.name)

    @staticmethod
    def list_solvers():
        """
        List all registered solvers
        """
        return SOLVER_REGISTRY
    
    @staticmethod
    def list_claimprocessors():
        """
        List all registered claim processors
        """
        # Get all claim processors
        claimprocessors = {}
        for solver, value in SOLVER_REGISTRY.items():
            if "claimprocessor" in solver:
                claimprocessors[solver] = value

        return claimprocessors
    
    @staticmethod
    def list_retrievers():
        """
        List all registered retrievers
        """
        # Get all retrievers
        retrievers = {}
        for solver, value in SOLVER_REGISTRY.items():
            if "retriever" in solver:
                retrievers[solver] = value

        return retrievers
    
    @staticmethod
    def list_verifiers():
        """
        List all registered verifiers
        """
        # Get all verifiers
        verifiers = {}
        for solver, value in SOLVER_REGISTRY.items():
            if "verifier" in solver:
                verifiers[solver] = value

        return verifiers
    
    def init_solver(self, solver_name, args):
        """
        Initialize a solver with the given configuration
        """

        # Check if the solver is registered
        if solver_name not in SOLVER_REGISTRY:
            logger.error(f"{solver_name} not in SOLVER_REGISTRY")
            raise RuntimeError(f"{solver_name} not in SOLVER_REGISTRY")
        
        # Initialize the solver
        solver_cls = SOLVER_REGISTRY[solver_name]
        solver_cls.input_name = args.get("input_name", solver_cls.input_name)
        solver_cls.output_name = args.get("output_name", solver_cls.output_name)

        logger.info(f"Solver {solver_cls(args)} initialized")

        return solver_cls(args), solver_cls.input_name, solver_cls.output_name
    
    def init_solvers(self):
        """
        Initialize all registered solvers
        """
        solvers = {}
        for k, v in self.solver_configs.items():
            solver, input_name, output_name = self.init_solver(k, v)
            solvers[k] = (solver, input_name, output_name)
        return solvers
    
    def init_pipeline(self):
        """
        Initialize the pipeline with the given configuration
        """
        pipeline = {}
        for required_solver in self.config.pipeline:
            if required_solver not in self.solver_configs:
                logger.error(f"{required_solver} not in solvers config")
                raise RuntimeError(f"{required_solver} not in solvers config")
            solver, input_name, output_name = self.init_solver(required_solver, self.solver_configs[required_solver])
            pipeline[required_solver] = (solver, input_name, output_name)

        return pipeline
    
    def init_pipeline_manually(self, pipeline: list):
        """
        Initialize the pipeline with the given configuration

        Parameters
        ----------
        pipeline : list
            A list of solvers to be included in the pipeline
        """
        self.pipeline = {}
        for required_solver in pipeline:
            if required_solver not in self.solver_configs:
                raise RuntimeError(f"{required_solver} not in solvers config")
            solver, input_name, output_name = self.init_solver(required_solver, self.solver_configs[required_solver])
            self.pipeline[required_solver] = (solver, input_name, output_name)

    def persist_output(self, state: FactCheckerState, idx, solver_name, cont, sample_name=0):
        result = {
            "idx": idx,
            "solver": solver_name,
            "continue": cont,
            "state": state.to_dict()
        }
        with open(os.path.join(self.output_path, f'{sample_name}.jsonl'), 'a', encoding="utf-8") as f:
            f.write(json.dumps(result, ensure_ascii=False) + '\n')

    def read_output(self, sample_name):
        """
        Read the output file for the given sample
        """
        with open(os.path.join(self.output_path, f'{sample_name}.jsonl'), 'r', encoding="utf-8") as f:
            return [json.loads(line) for line in f]
        
    def remove_output(self, sample_name):
        """
        Remove the output file for the given sample
        """
        os.remove(os.path.join(self.output_path, f'{sample_name}.jsonl'))

    def __call__(self, response: str, question: str = None, stream: bool = False, callback: Callable = None, **kwargs):
        """
        Evaluate the response using the pipeline
        """

        def evaluate_response():
            # Check if sample_name is provided in kwargs else generate a random one
            sample_name = kwargs.get("sample_name", str(uuid.uuid4().hex[:6]))

            # Initialize the state
            solver_output = FactCheckerState(question=question, response=response)

            # Initialize the output name
            output_name = "response"
            for idx, (name, (solver, input_name, output_name)) in tqdm.tqdm(enumerate(self.pipeline.items()),
                                                                total=len(self.pipeline)):
                logger.info(f"Invoking solver: {idx}-{name}")
                logger.info(f"State content: {solver_output}")
            
                try:
                    # Solver input is the output of the previous solver
                    solver_input = solver_output

                    # Run the solver
                    cont, solver_output = solver(solver_input, **kwargs)

                    # Persist the output
                    logger.debug(f"Latest result: {solver_output}")
                    if callback:
                        callback(
                            index=idx,
                            sample_name=sample_name,
                            solver_name=name,
                            input_name=input_name,
                            output_name=output_name,
                            input=solver_input.__dict__,
                            output=solver_output.__dict__,
                            continue_run=cont
                        )
                    
                    # Stream the output
                    if stream:
                        yield {
                            "index": idx,
                            "solver_name": name,
                            "input_name": input_name,
                            "output_name": output_name,
                            "input": solver_input.__dict__,
                            "output": solver_output.__dict__,
                            "continue_run": cont
                        }

                    self.persist_output(solver_output, idx, name, cont, sample_name=sample_name)
                    
                except:
                    logger.error(f"Error at {traceback.format_exc()}")
                    cont = False
                    output_name = input_name

                # Break if the solver returns False
                if not cont:
                    logger.info(f"Break at {name}")
                    break
                
            if not stream:
                return solver_output.get(output_name)
        
        # Execute the generator if stream is True, otherwise process normally
        return evaluate_response()


import argparse

from openfactcheck.core.base import OpenFactCheck
from openfactcheck.lib.config import OpenFactCheckConfig

def parse_args():
    parser = argparse.ArgumentParser(description='Initialize OpenFactCheck with custom configuration.')
    
    # Add arguments here, example:
    parser.add_argument("--config-path", 
                        type=str, 
                        help="Config File Path",
                        default="config.json")
    
    # Parse arguments from command line
    args = parser.parse_args()
    return args

if __name__ == "__main__":
    args = parse_args()

    ofc = OpenFactCheck(OpenFactCheckConfig(args.config_path))


import logging

from core.fact_check_state import FactCheckerState
from core.task_solver import StandardTaskSolver
from core import register_solver
import random
import string
from .rarr_utils.hallucination import run_evidence_hallucination
from .prompts.hallucination_prompts import EVIDENCE_HALLUCINATION


@register_solver("llm_retriever", "claims_with_questions", "claims_with_evidences")
class LLMRetriever(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.model = self.global_config.get("model", "gpt-3.5-turbo-instruct")

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claims = state.get(self.input_name)

        for claim, contents in claims.items():
            questions = contents.get("questions", [])
            evidences = []
            for question in questions:
                evidences.append(
                    run_evidence_hallucination(
                        question,
                        model=self.model,
                        prompt=EVIDENCE_HALLUCINATION
                    )
                )
            claims[claim]['evidences'] = evidences

        state.set(self.output_name, claims)
        return True, state



import re
import time
import pandas as pd
import streamlit as st

from openfactcheck.core.base import OpenFactCheck
from openfactcheck.app.utils import metric_card

def extract_text(claim):
    """
    Extracts text from a claim that might be a string formatted as a dictionary.
    """
    # Try to extract text using regular expression if claim is a string formatted as a dictionary
    match = re.search(r"'text': '([^']+)'", claim)
    if match:
        return match.group(1)
    return claim  # Return as is if no dictionary format detected

# Create a function to check a LLM response
def evaluate_response(ofc: OpenFactCheck):
    """
    This function creates a Streamlit app to evaluate the factuality of a LLM response.
    """

    # Initialize the solvers
    st.session_state.claimprocessors = ofc.list_claimprocessors()
    st.session_state.retrievers = ofc.list_retrievers()
    st.session_state.verifiers = ofc.list_verifiers()

    st.write("This is where you can check factuality of a LLM response.")

    # Customize FactChecker
    st.write("Customize FactChecker")

    # Dropdown in three columns
    col1, col2, col3 = st.columns(3)
    with col1:
        if "claimprocessor" not in st.session_state:
            st.session_state.claimprocessor = st.selectbox("Select Claim Processor", list(st.session_state.claimprocessors))
        else:
            st.session_state.claimprocessor = st.selectbox("Select Claim Processor", list(st.session_state.claimprocessors), index=list(st.session_state.claimprocessors).index(st.session_state.claimprocessor))
    with col2:
        if "retriever" not in st.session_state:
            st.session_state.retriever = st.selectbox("Select Retriever", list(st.session_state.retrievers))
        else:
            st.session_state.retriever = st.selectbox("Select Retriever", list(st.session_state.retrievers), index=list(st.session_state.retrievers).index(st.session_state.retriever))
    with col3:
        if "verifier" not in st.session_state:
            st.session_state.verifier = st.selectbox("Select Verifier", list(st.session_state.verifiers))
        else:
            st.session_state.verifier = st.selectbox("Select Verifier", list(st.session_state.verifiers), index=list(st.session_state.verifiers).index(st.session_state.verifier))

    # Input
    if "input_text" not in st.session_state:
        st.session_state.input_text = {"text": st.text_area("Enter LLM response here", "This is a sample LLM response.")}
    else:
        st.session_state.input_text = {"text": st.text_area("Enter LLM response here", st.session_state.input_text["text"])}

    # Button to check factuality
    if st.button("Check Factuality"):
        with st.status("Checking factuality...", expanded=True) as status:
            # Configure the pipeline
            st.write("Configuring pipeline...")
            ofc.init_pipeline_manually([st.session_state.claimprocessor, st.session_state.retriever, st.session_state.verifier])
            st.write("Pipeline configured...")

            # Evaluate the response
            st.write("Evaluating response...")

            response = ofc(st.session_state.input_text, stream=True)
            st.write("Response evaluated...")

            status.update(label="Factuality checked...", state="complete", expanded=False)

        # Display pipeline configuration
        pipeline_str = "&nbsp;&nbsp;&nbsp;┈➤&nbsp;&nbsp;&nbsp;".join([st.session_state.claimprocessor, st.session_state.retriever, st.session_state.verifier])
        st.info(f"""**Pipeline**:&nbsp;&nbsp;&nbsp; \n{pipeline_str}""")

        # Store the final response in the session state
        st.session_state.final_response = None

        col1, col2 = st.columns([3, 1])
        with col1:
            def process_stream(responses):
                """
                Process each response from the stream as a simulated chat output.
                This function yields each word from the formatted text of the response,
                adding a slight delay to simulate typing in a chat.
                """

                for response in responses:
                    if "claimprocessor" in response["solver_name"]:
                        # Extract response details
                        output_text = response["output"]

                        # Get the number of detected claims
                        detected_claims = output_text.get("claims", [])

                        # Generate formatted text with enumerated claims in Markdown format
                        formatted_text = "### Detected Claims\n"
                        formatted_text += "\n".join(f"{i}. {extract_text(claim)}" for i, claim in enumerate(detected_claims, start=1))
                        formatted_text += "\n"

                        with col2:
                            metric_card(label="Detected Claims", value=len(detected_claims))

                        # Yield each word with a space and simulate typing by sleeping
                        for word in formatted_text.split(" "):
                            yield word + " "
                            time.sleep(0.01)

                        st.session_state.claimprocessor_flag = True

                    elif "retriever" in response["solver_name"]:
                        # Extract response details
                        output_text = response["output"]

                        evidences = []
                        for _, claim_with_evidences in output_text.get("claims_with_evidences", {}).items():
                            for evidence in claim_with_evidences:
                                evidences.append(evidence[1])

                        # # Generate formatted text with enumerated evidences in Markdown format
                        # formatted_text = "#### Retrieved Evidences\n"
                        # formatted_text += "\n".join(f"{i}. {evidence}" for i, evidence in enumerate(evidences, start=1))
                        # formatted_text += "\n"

                        with col2:
                            metric_card(label="Retrieved Evidences", value=len(evidences))

                        # # Yield each word with a space and simulate typing by sleeping
                        # for word in formatted_text.split(" "):
                        #     yield word + " "
                        #     time.sleep(0.01)

                    elif "verifier" in response["solver_name"]:
                        # Extract response details
                        output_text = response["output"]

                        # Get detail
                        details = output_text.get("detail", None)
                        if details is None:
                            detail_text = "The verifier did not provide any detail. Please use other verifiers for more information."
                        else:
                            detail_text = ""

                            # Apply color to the claim based on factuality
                            claims=0
                            false_claims = 0
                            true_claims = 0
                            for i, detail in enumerate(details):
                                if detail.get("factuality", None) is not None:
                                    claim=detail.get("claim", "")
                                    if detail.get("factuality", None) == -1:
                                        detail_text += f'##### :red[{str(i+1) + ". " + extract_text(claim)}]'
                                        detail_text += "\n"
                                        claims += 1
                                        false_claims += 1
                                    elif detail.get("factuality", None) == 1:
                                        detail_text += f'##### :green[{str(i+1) + ". " + extract_text(claim)}]'
                                        detail_text += "\n"
                                        claims += 1
                                        true_claims += 1
                                    else:
                                        detail_text += f'##### :yellow[{str(i+1) + ". " + extract_text(claim)}]'
                                        detail_text += "\n"
                                        claims += 1
                                else:
                                    st.error("Factuality not found in the verifier output.")

                                # Add error information
                                if detail.get("error", None) is not "None":
                                    detail_text += f"- **Error**: {detail.get('error', '')}"
                                    detail_text += "\n"

                                # Add reasoning information
                                if detail.get("reasoning", None) is not "None":
                                    detail_text += f"- **Reasoning**: {detail.get('reasoning', '')}"
                                    detail_text += "\n"
                                
                                # Add correction
                                if detail.get("correction", None) is not "":
                                    detail_text += f"- **Correction**: {detail.get('correction', '')}"
                                    detail_text += "\n"

                                # Add evidence
                                if detail.get("evidence", None) is not "":
                                    evidence_text = ""
                                    for evidence in detail.get("evidences", []):
                                        evidence_text += f"  - {evidence[1]}"
                                        evidence_text += "\n"
                                    detail_text += f"- **Evidence**:\n{evidence_text}"

                                        
                        # Generate formatted text with the overall factuality in Markdown format
                        formatted_text = "### Factuality Detail\n"
                        formatted_text += "Factuality of each claim is color-coded (red:[red means false], green:[green means true], yellow:[yellow means unknown]) as follows:\n"
                        formatted_text += f"{detail_text}\n"
                        formatted_text += "\n"

                        # Get the number of true and false claims
                        with col2:
                            metric_card(label="Supported Claims", value=true_claims, background_color="#D1ECF1", border_left_color="#17A2B8")
                            metric_card(label="Conflicted Claims", value=false_claims, background_color="#D1ECF1", border_left_color="#17A2B8")
                        
                        # Get overall factuality (label)
                        overall_factuality = output_text.get("label", "Unknown")
                        with col2:
                            with st.container():
                                if overall_factuality == True:
                                    metric_card(label="Overall Factuality", value="True", background_color="#D4EDDA", border_left_color="#28A745")
                                elif overall_factuality == False:
                                    metric_card(label="Overall Factuality", value="False", background_color="#F8D7DA", border_left_color="#DC3545")

                        # Get overall credibility (score)
                        overall_credibility = true_claims / claims if claims > 0 else 0
                        with col2:
                            if overall_credibility > 0.75 and overall_credibility <= 1:
                                # Green background
                                metric_card(label="Overall Credibility", value=f"{overall_credibility:.2%}", background_color="#D4EDDA", border_left_color="#28A745")
                            elif overall_credibility > 0.25 and overall_credibility <= 0.75:
                                # Yellow background
                                metric_card(label="Overall Credibility", value=f"{overall_credibility:.2%}", background_color="#FFF3CD", border_left_color="#FFC107")
                            else:
                                # Red background
                                metric_card(label="Overall Credibility", value=f"{overall_credibility:.2%}", background_color="#F8D7DA", border_left_color="#DC3545")

                        # Yield each word with a space and simulate typing by sleeping
                        for word in formatted_text.split(" "):
                            yield word + " "
                            time.sleep(0.01)

            st.write_stream(process_stream(response))
                            


from core import register_solver, FactCheckerState, StandardTaskSolver

##
#
# Factool Data Post-Editor
#
# Notes:
#   Factool response post-processor. Used to presents the results in human-readable format and to save the analysis in a JSON file.
#
##
@register_solver("factool_post_editor", "claim_info", "claim_info")
class FactoolPostEditor(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claim_info = state.get(self.input_name)

        for key, pair in claim_info.items():
            claim_info[key]['edited_claims'] = claim_info[key]['stances'][0]

        state.set(self.output_name, claim_info)
        return True, state



import os
from pipeline import Pipeline
from argparse import Namespace

# Base directory where the script is located
base_dir = os.path.abspath(os.path.dirname(__file__))

args = Namespace(
    user_src=os.path.join(base_dir),
    config=os.path.join(base_dir, "../../config/factool_config.yaml"),
    output=os.path.join(base_dir, "../../../output")
)

p = Pipeline(args)
question = "Who is Alan Turing?"
response = "Alan Turing was a British mathematician, logician, cryptanalyst, and computer scientist. He was highly influential in the development of theoretical computer science."

print(p(question=question, response=response))



from openfactcheck.core.state import FactCheckerState
from openfactcheck.core.solver import StandardTaskSolver, Solver

from .rarr_utils.agreement_gate import run_agreement_gate
from .rarr_utils.functional_prompt import AGREEMENT_GATE_PROMPT


@Solver.register("rarr_verifier", "claims_with_evidences", "label")
class RARRAgreementGate(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.max_evidences_per_question = args.get("max_evidences_per_question", 1)
        self.model = self.global_config.get("rarr_model", "gpt-3.5-turbo-instruct")

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claims_with_evidences = state.get(self.input_name)
        results = []
        for claim, evidences in claims_with_evidences.items():
            result = {}
            evidences = evidences[:self.max_evidences_per_question]
            labels = []
            for query, evidence in evidences:
                gate = run_agreement_gate(
                    claim=claim,
                    context=None,
                    query=query,
                    evidence=evidence,
                    model=self.model,
                    prompt=AGREEMENT_GATE_PROMPT
                )
                labels.append(gate['is_open'])
            result['claim'] = claim
            result['evidences'] = evidences
            result['labels'] = labels
            result['factuality'] = all(labels)
            results.append(result)
        state.set(self.output_name, all([x['factuality'] for x in results]))
        state.set("detail", results)
        return True, state



from core import register_solver, StandardTaskSolver, FactCheckerState
import asyncio
import nest_asyncio
from factool import Factool
from .ftool_utils.chat_api import OpenAIChat
import yaml
import os
from typing import List


##
#
# Factool Decontextualizer
#
# Notes:
#   - This solver is used to extract claims from a response.
#   - The response should be a string.
#
##
@register_solver("factool_decontextualizer", "response", "claims")
class FactoolDecontextualizer(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.gpt_model = self.global_config.get("llm_in_use", "gpt-3.5-turbo")
        self.gpt = OpenAIChat(self.gpt_model)
        self.claim_prompt = yaml.load(
            open(
                os.path.join(
                    os.path.dirname(os.path.abspath(__file__)),
                    "ftool_utils/prompts.yaml",
                ),
                "r",
            ),
            yaml.FullLoader,
        )["claim_extraction"]

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        response = state.get(self.input_name)

        claims = self._claim_extraction(responses=[response])[0]

        extracted_claims = [claim["claim"] for claim in claims]

        state.set(self.output_name, extracted_claims)
        return True, state

    def _claim_extraction(self, responses):
        messages_list = [
            [
                {"role": "system", "content": self.claim_prompt["system"]},
                {
                    "role": "user",
                    "content": self.claim_prompt["user"].format(input=response),
                },
            ]
            for response in responses
        ]
        return self.gpt.run(messages_list, List)



from core import register_solver, FactCheckerState, StandardTaskSolver
from factool import Factool
import os


##
#
# Factool Solver
#
# Notes:
# Factool requires 3 input parameters: prompt, response, and category.
# Category is always set to 'kbqa' (Knowledge Base Question Answering) for the purposes of this project.
# Because of employing a pipeline of its own, with specific search engine and analysis tools, Factool requires several API keys to be set as environment variables.
# That is:
#   openai_key - OpenAI API key (https://beta.openai.com/)
#   serper_key - Serper API key (https://serper.dev/)
#   scrapper_key - Scrapper API key (https://www.scraperapi.com/)
# Additional parameters:
#   llm_in_use - The OpenAI LLM in use (e.g. gpt-4)
#
##
@register_solver("factool_blackbox", "response", "claim_info")
class FactoolBlackboxSolver(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.input_prompt = args.get("input_prompt", None)
        self.gpt_model = self.global_config.get("llm_in_use", "gpt-4")
        # self.input_prompt = args["input_prompt"] if "input_prompt" in args else None
        # self.gpt_model = args["llm_in_use"] if "llm_in_use" in args else "gpt-4"

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        prompt = state.get(self.input_prompt)
        response = state.get(self.input_name)

        factool_instance = Factool(self.gpt_model)

        inputs = [{"prompt": prompt, "response": response, "category": "kbqa"}]
        claim_info = factool_instance.run(inputs)

        state.set("claim_info", claim_info)
        return True, state



import os
import json
import yaml
import datasets
import transformers
from pathlib import Path
from typing import Union
from collections import namedtuple
from importlib import resources as pkg_resources

from openfactcheck.lib.logger import logger, set_logger_level
from openfactcheck.lib.errors import ConfigValidationError
from openfactcheck import templates as solver_config_templates_dir
from openfactcheck import solvers as solver_templates_dir

# Import solver configuration templates
solver_config_templates_path = pkg_resources.files(solver_config_templates_dir) / 'solver_configs'
with solver_config_templates_path as solver_config_templates_dir_path:
    solver_config_template_files = [str(f) for f in solver_config_templates_dir_path.iterdir()]

# Import default solvers
# TODO: Currently, only webservice solvers are supported as default solvers
solver_templates_paths = [
    str(pkg_resources.files(solver_templates_dir) / 'webservice')
]

class OpenFactCheckConfig:
    """
    Class to load the OpenFactCheck configuration from a JSON or YAML file.

    Parameters
    ----------
    filename: str, Path
        The path to the configuration file.

    Attributes
    ----------
    retries: int
        Number of retries for the pipeline components.
    pipeline: namedtuple
        Namedtuple containing the pipeline components.
    solver_configs: dict
        Dictionary containing the solver configurations.
    solver_paths: dict
        Dictionary containing the paths to the solver models.
    output_path: str
        Path to the output directory.
    secrets: namedtuple
        Namedtuple containing the API keys.
    verbose: str
        The verbosity level for the logger.

    Methods
    -------
    solver_configuration(solver: str = None) -> dict:
        Get the solver configuration for a specific solver or all solvers.
    validate():
        Validate the configuration file

    Examples
    --------
    >>> config = OpenFactCheckConfig("config.json")
    """
    def __init__(self, filename: Union[str, Path] = "config.json"):
        # Setup Logger
        self.logger = logger
        self.filename = filename

        # Define namedtuple structures
        Secrets = namedtuple("Secrets", ["openai_api_key", 
                                         "serper_api_key", 
                                         "azure_search_key"])
        Pipeline = namedtuple("Pipeline", ["claimprocessor",
                                           "retriever",
                                           "verifier"])
        
        # Define Attributes
        self.config = None
        self.retries = 0
        self.pipeline = None
        self.solver_configs = None
        self.solver_paths = None
        self.output_path = None
        self.secrets = None
        self.verbose = ""

        try:
            # Check if the file exists
            if Path(self.filename).exists():
                # Loading Config File
                with open(self.filename, encoding="utf-8") as file:
                    self.config = json.load(file)
                    self.logger.info(f"Config file loaded successfully from {self.filename}")
            else:
                # Create a dummy configuration file
                self.logger.warning(f"Config file not found: {self.filename}")
                self.config = {}

            # Initialize Retries
            if 'retries' in self.config:
                self.retries = self.config['retries']
            else:
                self.logger.warning("No retries found in the configuration file. Using default value of 3.")
                self.retries = 3
                
            # Initialize template solvers along with the user-defined solvers
            # User defined solvers will override the template solvers
            if 'solver_configs' in self.config:
                self.solver_configs = SolversConfig(solver_config_template_files + self.config['solver_configs'])()
            else:
                self.logger.warning("No solver configurations found in the configuration file. Using default templates only.")
                self.solver_configs = SolversConfig(solver_config_template_files)()

            # Initialize template solver paths along with the user-defined solver paths
            if 'solver_paths' in self.config:
                self.solver_paths = solver_templates_paths + self.config['solver_paths']
            else:
                self.logger.warning("No solver paths found in the configuration file. Using default solver paths only.")
                self.solver_paths = solver_templates_paths

            # Initialize Output Path
            if 'output_path' in self.config:
                self.output_path = self.config['output_path']
                os.makedirs(self.output_path, exist_ok=True)
            else:
                self.logger.warning("No output path found in the configuration file. Using default output path 'tmp/output'.")
                self.output_path = "tmp/output"
                os.makedirs(self.output_path, exist_ok=True)
            
            # Initialize Pipeline config
            if 'pipeline' in self.config:
                self.pipeline = Pipeline(claimprocessor=self.config['pipeline']['claimprocessor'],
                                            retriever=self.config['pipeline']['retriever'],
                                            verifier=self.config['pipeline']['verifier'])
            else:
                if self.solver_configs:
                    solvers = list(self.solver_configs.keys())
                    claimprocessor = None
                    retriever = None
                    verifier = None
                    for solver in solvers:
                        if 'claimprocessor' in solver:
                            claimprocessor = solver
                        if 'retriever' in solver:
                            retriever = solver
                        if 'verifier' in solver:
                            verifier = solver
                        if claimprocessor and retriever and verifier:
                            break
                    self.pipeline = Pipeline(claimprocessor=claimprocessor, retriever=retriever, verifier=verifier)
                    self.logger.warning(f"No pipeline found in the configuration file. Using first solver as default pipeline. ClaimProcessor: {claimprocessor}, Retriever: {retriever}, Verifier: {verifier}")

            # Initialize Secrets config
            if 'secrets' in self.config:
                self.secrets = Secrets(openai_api_key=self.config['secrets']['openai_api_key'],
                                        serper_api_key=self.config['secrets']['serper_api_key'],
                                        azure_search_key=self.config['secrets']['azure_search_key'])
            else:
                self.logger.warning("No secrets found in the configuration file. Make sure to set the environment variables.")
                self.secrets = Secrets(openai_api_key=None, serper_api_key=None, azure_search_key=None)
        
            # Initialize Environment Variables
            if self.secrets.openai_api_key:
                os.environ['OPENAI_API_KEY'] = self.secrets.openai_api_key
            if self.secrets.serper_api_key:
                os.environ['SERPER_API_KEY'] = self.secrets.serper_api_key
            if self.secrets.azure_search_key:
                os.environ['AZURE_SEARCH_KEY'] = self.secrets.azure_search_key

            # Initialize Verbose
            if 'verbose' in self.config:
                self.verbose = self.config['verbose']
                set_logger_level(self.logger, self.verbose)
            else:
                self.logger.warning("No verbose level found in the configuration file. Using default level 'INFO'.")
                self.verbose = "INFO"
                set_logger_level(self.logger, "INFO")

            # Validate Configuration
            self.validate()

            # Disable Transformers and Datasets logging
            transformers.logging.set_verbosity_error()
            datasets.logging.set_verbosity_error()

        except FileNotFoundError:
            self.logger.error(f"Config file not found: {self.filename}")
            raise FileNotFoundError(f"Config file not found: {self.filename}")
        
        except json.JSONDecodeError:
            self.logger.error(f"Invalid JSON in config file: {self.filename}")
            raise ValueError(f"Invalid JSON in config file: {self.filename}")
        
        except ConfigValidationError as e:
            self.logger.error(f"Configuration validation failed: {e}")
            raise ConfigValidationError(f"Configuration validation failed: {e}")
        
        except Exception as e:
            self.logger.error(f"Unexpected error loading config file: {e}")
            raise Exception(f"Unexpected error loading config file: {e}")
        
    def validate(self):
        """
        Validate the configuration file.

        Raises
        ------
        ValueError
            If the configuration file is invalid.

        Examples
        --------
        >>> config = OpenFactCheckConfig("config.json")
        >>> config.validate()
        """
        # Check for environment variables
        if 'OPENAI_API_KEY' not in os.environ:
            self.logger.warning("OPENAI_API_KEY environment variable not found.")
            raise ConfigValidationError("OPENAI_API_KEY environment variable not found.")
        if 'SERPER_API_KEY' not in os.environ:
            self.logger.warning("SERPER_API_KEY environment variable not found.")
            raise ConfigValidationError("SERPER_API_KEY environment variable not found.")
        if 'AZURE_SEARCH_KEY' not in os.environ:
            self.logger.warning("AZURE_SEARCH_KEY environment variable not found.")
            raise ConfigValidationError("AZURE_SEARCH_KEY environment variable not found.")

        
    
    def solver_configuration(self, solver: str = None) -> dict:
        """
        Get the solver configuration for a specific solver or all solvers.

        Parameters
        ----------
        solver: str
            The name of the solver to get the configuration for.
            If not provided, returns the configuration for all solvers.

        Returns
        -------
        dict
            The configuration for the specified solver or all solvers.

        Raises
        ------
        ConfigValidationError
            If the configuration validation fails.

        Examples
        --------
        >>> config = OpenFactCheckConfig("config.json")
        >>> config.solver_configuration()
        >>> config.solver_configuration("factcheckgpt_claimprocessor")
        """
        if solver:
            if solver in self.solver_configs:
                return self.solver_configs[solver]
            else:
                self.logger.error(f"Solver not found: {solver}")
                raise ValueError(f"Solver not found: {solver}")
        else:
            return self.solver_configs
                

class SolversConfig:
    """
    Class to load the solvers configuration from one or more JSON or YAML files.
    Merges all configurations into a single dictionary.

    Parameters
    ----------
    filename(s): str, Path, list
        The path to the solvers configuration or a list of paths to multiple solvers configurations.
    """
    def __init__(self, filename_s: Union[str, Path, list]):
        self.logger = logger
        self.filename_s = filename_s
        self.solvers = {}

        try:
            if isinstance(self.filename_s, (str, Path)):
                self.load_config(self.filename_s)
            elif isinstance(self.filename_s, list):
                for filename in self.filename_s:
                    self.load_config(filename)
            else:
                self.logger.error(f"Invalid filename type: {type(self.filename_s)}")
                raise ValueError(f"Invalid filename type: {type(self.filename_s)}")
            
        except FileNotFoundError:
            self.logger.error(f"Solvers file not found: {self.filename_s}")
            raise FileNotFoundError(f"Solvers file not found: {self.filename_s}")
        except json.JSONDecodeError:
            self.logger.error(f"Invalid JSON in solvers file: {self.filename_s}")
            raise ValueError(f"Invalid JSON in solvers file: {self.filename_s}")
        except Exception as e:
            self.logger.error(f"Unexpected error loading solvers file: {e}")
            raise Exception(f"Unexpected error loading solvers file: {e}")

    def load_config(self, filename: Union[str, Path]):
        with open(filename, encoding="utf-8") as file:
            if filename.endswith(".yaml"):
                file_data = yaml.load(file, Loader=yaml.FullLoader) 
            elif filename.endswith(".json"):
                file_data = json.load(file)
            else:
                self.logger.error(f"Invalid file format: {filename}")
                raise ValueError(f"Invalid file format: {filename}")

            # Merge current file data into existing solvers dictionary
            self.solvers.update(file_data)

            # Log the loaded configuration pattern
            if 'template' in filename:
                self.logger.info(f"Template solver configuration loaded: {filename.split('/')[-1]}")
            else:
                self.logger.info(f"User-defined solver configuration loaded from: {filename}")

    def __call__(self):
        return self.solvers


import logging

from core.fact_check_state import FactCheckerState
from core.task_solver import StandardTaskSolver
from core import register_solver


@register_solver("all_pass_abstain_detector", "response", "response")
class AllPassAbstainDetector(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        return True, state



import json
from core import register_solver, FactCheckerState, StandardTaskSolver

# Factool example.py category=kbqa response protocol
'''
{
 'average_claim_level_factuality': 0.0,
 'average_response_level_factuality': 0.0,
 'detailed_information': [
    {
      'prompt': 'Introduce Graham Neubig',
      'response': 'Graham Neubig is a professor at MIT',
      'category': 'kbqa',
      'claims': [
        {
            'claim': 'Graham Neubig is a professor at MIT'
        }
      ],
      'queries': [
        [ 'Is Graham Neubig a professor at MIT?', 'Graham Neubig professorship' ]
      ],
      'evidences': [
        {
            'evidence': [ 'I am an Associate Professor at the Carnegie Mellon University Language Technology Institute in the School of Computer Science, and work with a bunch of great ...', 'Missing: MIT? | Show results with:MIT?', 'EI Seminar - Graham Neubig - Learning to Explain and ...', 'Duration: 57:54', 'Posted: Feb 17, 2023', 'I am an Associate Professor at the Carnegie Mellon University Language Technology Institute in the School of Computer Science, and work with a bunch of great ...', 'My research is concerned with language and its role in human communication. In particular, my long-term research goal is to break down barriers in human-human ...', 'Graham Neubig. Associate Professor. Research Interests: Machine Translation · Natural Language Processing · Spoken Language Processing · Machine Learning. My ...', "I am an Associate Professor of Computer Science at Carnegie Mellon University and CEO of… | Learn more about Graham Neubig's work experience, education, ...", 'Graham Neubig received the B.E. degree from the University of Illinois, Urbana ... He is currently an Assistant Professor with Carnegie Mellon University ...' ],
            'source': [ 'http://www.phontron.com/', 'http://www.phontron.com/', 'https://youtube.com/watch?v=CtcP5bvODzY', 'https://youtube.com/watch?v=CtcP5bvODzY', 'https://youtube.com/watch?v=CtcP5bvODzY', 'http://www.phontron.com/', 'https://www.phontron.com/research.php', 'https://lti.cs.cmu.edu/people/222217661/graham-neubig', 'https://www.linkedin.com/in/graham-neubig-10b41616b', 'https://ieeexplore.ieee.org/author/37591106000' ]
        }
      ],
      'claim_level_factuality': [
        {
            'reasoning': 'The given text is non-factual. Multiple pieces of evidence indicate that Graham Neubig is an Associate Professor at the Carnegie Mellon University Language Technology Institute in the School of Computer Science, not at MIT.',
            'error': 'Graham Neubig is not a professor at MIT.',
            'correction': 'Graham Neubig is a professor at Carnegie Mellon University.',
            'factuality': False,
            'claim': 'Graham Neubig is a professor at MIT'
        }
      ],
      'response_level_factuality': False
    }
  ]
}
'''

##
#
# Factool Data Post-Editor
#
# Notes:
#   Factool response post-processor. Used to presents the results in human-readable format and to save the analysis in a JSON file.
#
##
@register_solver("factool_blackbox_post_editor", "claim_info", "claim_info")
class FactoolBlackboxPostEditor(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.path_save_analysis = args.get("path_save_analysis","factool_evidence_analysis.json")

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claim_info = state.get(self.input_name)

        # Restructure some of the output for concatenation (corrected claims)
        edited_claims = ''
        for clf in claim_info['detailed_information'][0]['claim_level_factuality']:
            edited_claims += 'Claim: "' + clf['claim'] + '" => '
            edited_claims += ('' if (clf['error'] == 'None' or len(clf['error']) == 0) else (clf['error'] + ' '))
            edited_claims += ('' if (clf['reasoning'] == 'None' or len(clf['reasoning']) == 0) else clf['reasoning'])
            edited_claims += ((' ' + clf['claim']) if (clf['correction'] == 'None' or len(clf['correction']) == 0) else (' ' + clf['correction']))
            edited_claims += '\n'
        edited_claims = edited_claims[:-1]
        new_claim_info = {}
        new_claim_info[claim_info['detailed_information'][0]['response']] = {
            "edited_claims": edited_claims
        }

        # Serializing json
        json_object = json.dumps(claim_info, indent=4)

        # Writing to sample.json
        with open(self.path_save_analysis, "w") as outfile:
            outfile.write(json_object)

        state.set(self.output_name, new_claim_info)
        return True, state
